from __future__ import annotations

from collections.abc import Iterable
from io import BytesIO
from urllib.parse import quote_plus

from asserts import (
    assert_equal,
    assert_in,
    assert_not_in,
    assert_raises,
    assert_succeeds,
    fail,
)
from dectest import TestCase, before, test
from werkzeug.exceptions import BadRequest

from rouver.args import ArgumentParser, Multiplicity, parse_args
from rouver.exceptions import ArgumentsError
from rouver.util import rfc5987_encode
from rouver_test.testutil import default_environment

MULTIPART_PART_TMPL = """--1234567890
Content-Disposition: form-data; name="{name}"

{value}
"""

MULTIPART_FILE_BODY_TMPL = """--1234567890
Content-Disposition: form-data; name="{name}"; {filename_param}
Content-Type: {type}

{content}
--1234567890--
"""


class ParseArgsTest(TestCase):
    @before
    def setup_environment(self) -> None:
        self.env = default_environment()

    def add_path_argument(self, name: str, value: str) -> None:
        if "QUERY_STRING" not in self.env:
            self.env["QUERY_STRING"] = ""
        if self.env["QUERY_STRING"]:
            self.env["QUERY_STRING"] += "&"
        self.env["QUERY_STRING"] += "{}={}".format(
            quote_plus(name), quote_plus(value)
        )

    def setup_empty_urlencoded_request(self) -> None:
        self.env[
            "CONTENT_TYPE"
        ] = "application/x-www-form-urlencoded; charset=utf-8"

    def setup_urlencoded_request(self, name: str, value: str) -> None:
        self.setup_empty_urlencoded_request()
        body = "{}={}".format(name, value).encode("utf-8")
        self.env["CONTENT_LENGTH"] = str(len(body))
        self.env["wsgi.input"] = BytesIO(body)

    def setup_multipart_request(
        self, name: str, value: str | Iterable[str]
    ) -> None:
        self.env["CONTENT_TYPE"] = "multipart/form-data; boundary=1234567890"
        if isinstance(value, str):
            value = [value]
        body = (
            "".join(
                MULTIPART_PART_TMPL.format(name=name, value=v) for v in value
            )
            + "--1234567890--"
        ).encode("utf-8")
        self.env["CONTENT_LENGTH"] = str(len(body))
        self.env["wsgi.input"] = BytesIO(body)

    def setup_multipart_file_request(
        self, name: str, filename: str, file_content: str, content_type: str
    ) -> None:
        self.env["CONTENT_TYPE"] = "multipart/form-data; boundary=1234567890"
        filename_param = rfc5987_encode("filename", filename)
        body = MULTIPART_FILE_BODY_TMPL.format(
            name=name,
            filename_param=filename_param,
            content=file_content,
            type=content_type,
        ).encode("utf-8")
        self.env["CONTENT_LENGTH"] = str(len(body))
        self.env["wsgi.input"] = BytesIO(body)

    @test
    def parse_nothing(self) -> None:
        args = parse_args(self.env, [])
        assert_equal({}, args)

    @test
    def invalid_value_parser(self) -> None:
        with assert_raises(TypeError):
            parse_args(
                self.env,
                [
                    (
                        "foo",  # type: ignore
                        "INVALID",
                        Multiplicity.OPTIONAL,
                    )
                ],
            )

    @test
    def parse_str_arg(self) -> None:
        self.add_path_argument("foo", "bar")
        args = parse_args(self.env, [("foo", str, Multiplicity.REQUIRED)])
        assert_equal({"foo": "bar"}, args)

    @test
    def parse_unicode_arg(self) -> None:
        self.add_path_argument("föo", "bär")
        args = parse_args(self.env, [("föo", str, Multiplicity.REQUIRED)])
        assert_equal({"föo": "bär"}, args)

    @test
    def parse_int_arg(self) -> None:
        self.add_path_argument("foo", "123")
        args = parse_args(self.env, [("foo", int, Multiplicity.REQUIRED)])
        assert_equal({"foo": 123}, args)

    @test
    def parse_invalid_int_arg(self) -> None:
        self.add_path_argument("foo", "bar")
        try:
            parse_args(self.env, [("foo", int, Multiplicity.REQUIRED)])
        except ArgumentsError as exc:
            assert_equal(
                {"foo": "invalid literal for int() with base 10: 'bar'"},
                exc.arguments,
            )
        else:
            fail("ArgumentsError not raised")

    @test
    def required_argument_missing(self) -> None:
        try:
            parse_args(self.env, [("foo", str, Multiplicity.REQUIRED)])
        except ArgumentsError as exc:
            assert_equal({"foo": "mandatory argument missing"}, exc.arguments)
        else:
            fail("ArgumentsError not raised")

    @test
    def optional_argument(self) -> None:
        self.add_path_argument("foo", "bar")
        args = parse_args(self.env, [("foo", str, Multiplicity.OPTIONAL)])
        assert_equal({"foo": "bar"}, args)

    @test
    def optional_argument_missing(self) -> None:
        args = parse_args(self.env, [("foo", str, Multiplicity.OPTIONAL)])
        assert_equal({}, args)

    @test
    def optional_argument_empty(self) -> None:
        self.add_path_argument("foo", "")
        args = parse_args(self.env, [("foo", str, Multiplicity.OPTIONAL)])
        assert_equal({"foo": ""}, args)

    @test
    def any_argument_empty(self) -> None:
        args = parse_args(self.env, [("foo", int, Multiplicity.ANY)])
        assert_equal({"foo": []}, args)

    @test
    def any_argument(self) -> None:
        self.add_path_argument("foo", "123")
        self.add_path_argument("foo", "456")
        self.add_path_argument("foo", "789")
        args = parse_args(self.env, [("foo", int, Multiplicity.ANY)])
        assert_equal({"foo": [123, 456, 789]}, args)

    @test
    def required_any_argument(self) -> None:
        self.add_path_argument("foo", "123")
        self.add_path_argument("foo", "456")
        self.add_path_argument("foo", "789")
        args = parse_args(self.env, [("foo", int, Multiplicity.REQUIRED_ANY)])
        assert_equal({"foo": [123, 456, 789]}, args)

    @test
    def required_any_argument_missing(self) -> None:
        try:
            parse_args(self.env, [("foo", int, Multiplicity.REQUIRED_ANY)])
        except ArgumentsError as exc:
            assert_equal({"foo": "mandatory argument missing"}, exc.arguments)
        else:
            fail("ArgumentsError not raised")

    @test
    def urlencoded_post_request(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_urlencoded_request("foo", "bar")
        args = parse_args(self.env, [("foo", str, Multiplicity.OPTIONAL)])
        assert_equal({"foo": "bar"}, args)

    @test
    def urlencoded_post_request_with_umlauts(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_urlencoded_request("föo", "bär")
        args = parse_args(self.env, [("föo", str, Multiplicity.OPTIONAL)])
        assert_equal({"föo": "bär"}, args)

    @test
    def urlencoded_patch_request(self) -> None:
        self.env["REQUEST_METHOD"] = "PATCH"
        self.setup_urlencoded_request("foo", "bar")
        args = parse_args(self.env, [("foo", str, Multiplicity.OPTIONAL)])
        assert_equal({"foo": "bar"}, args)

    @test
    def urlencoded_delete_request(self) -> None:
        self.env["REQUEST_METHOD"] = "DELETE"
        self.setup_urlencoded_request("foo", "bar")
        args = parse_args(self.env, [("foo", str, Multiplicity.OPTIONAL)])
        assert_equal({"foo": "bar"}, args)

    @test
    def empty_delete__optional(self) -> None:
        self.env["REQUEST_METHOD"] = "DELETE"
        self.setup_empty_urlencoded_request()
        args = parse_args(
            self.env,
            [
                ("opt", str, Multiplicity.OPTIONAL),
                ("any", str, Multiplicity.ANY),
            ],
        )
        assert_not_in("opt", args)
        assert_equal([], args["any"])

    @test
    def empty_delete__required_not_supplied(self) -> None:
        self.env["REQUEST_METHOD"] = "DELETE"
        self.setup_empty_urlencoded_request()
        with assert_raises(ArgumentsError):
            parse_args(
                self.env,
                [
                    ("req", str, Multiplicity.REQUIRED),
                    ("once", str, Multiplicity.REQUIRED_ANY),
                ],
            )

    @test
    def urlencoded_put_request(self) -> None:
        self.env["REQUEST_METHOD"] = "PUT"
        self.setup_urlencoded_request("foo", "bar")
        args = parse_args(self.env, [("foo", str, Multiplicity.OPTIONAL)])
        assert_equal({"foo": "bar"}, args)

    @test
    def urlencoded_put_request_with_umlauts(self) -> None:
        self.env["REQUEST_METHOD"] = "PUT"
        self.setup_urlencoded_request("föo", "bär")
        args = parse_args(self.env, [("föo", str, Multiplicity.OPTIONAL)])
        assert_equal({"föo": "bär"}, args)

    @test
    def multipart_post_request(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_multipart_request("foo", "bar")
        args = parse_args(self.env, [("foo", str, Multiplicity.OPTIONAL)])
        assert_equal({"foo": "bar"}, args)

    @test
    def multipart_post_request_with_umlauts(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_multipart_request("föo", "bär")
        args = parse_args(self.env, [("föo", str, Multiplicity.OPTIONAL)])
        assert_equal({"föo": "bär"}, args)

    @test
    def multipart_put_request(self) -> None:
        self.env["REQUEST_METHOD"] = "PUT"
        self.setup_multipart_request("foo", "bar")
        args = parse_args(self.env, [("foo", str, Multiplicity.OPTIONAL)])
        assert_equal({"foo": "bar"}, args)

    @test
    def multipart_put_request_with_umlauts(self) -> None:
        self.env["REQUEST_METHOD"] = "PUT"
        self.setup_multipart_request("föo", "bär")
        args = parse_args(self.env, [("föo", str, Multiplicity.OPTIONAL)])
        assert_equal({"föo": "bär"}, args)

    @test
    def multipart_multiple_arguments(self) -> None:
        self.env["REQUEST_METHOD"] = "PUT"
        self.setup_multipart_request("foo", ["bar", "baz"])
        args = parse_args(self.env, [("foo", str, Multiplicity.ANY)])
        assert_equal({"foo": ["bar", "baz"]}, args)

    @test
    def multipart__optional_argument_empty(self) -> None:
        self.env["REQUEST_METHOD"] = "PUT"
        self.setup_multipart_request("foo", "")
        args = parse_args(self.env, [("foo", str, Multiplicity.OPTIONAL)])
        assert_equal({"foo": ""}, args)

    @test
    def multipart_post_request_with_file(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_multipart_file_request(
            "my-file", "my-file.txt", "content", "text/plain; charset=us-ascii"
        )
        args = parse_args(
            self.env, [("my-file", "file", Multiplicity.REQUIRED)]
        )
        assert_in("my-file", args)
        f = args["my-file"]
        assert_equal("my-file.txt", f.filename)
        assert_equal("text/plain", f.content_type)
        assert_equal(b"content", f.read())

    @test
    def multipart_post_request_with_optional_file(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_multipart_file_request(
            "my-file", "my-file.txt", "content", "text/plain; charset=us-ascii"
        )
        args = parse_args(
            self.env, [("my-file", "file-or-str", Multiplicity.REQUIRED)]
        )
        assert_in("my-file", args)
        f = args["my-file"]
        assert_equal("my-file.txt", f.filename)
        assert_equal("text/plain", f.content_type)
        assert_equal(b"content", f.read())

    @test
    def multipart_post_request_with_empty_file(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_multipart_request("my-file", "test")
        args = parse_args(
            self.env, [("my-file", "file-or-str", Multiplicity.REQUIRED)]
        )
        assert_in("my-file", args)
        assert_equal("test", args["my-file"])

    @test
    def multipart_post_request_with_file_and_umlauts(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_multipart_file_request(
            "föo", "my-filé.txt", "cöntent", "text/plain; charset=utf-8"
        )
        args = parse_args(self.env, [("föo", "file", Multiplicity.REQUIRED)])
        assert_in("föo", args)
        f = args["föo"]
        assert_equal("my-filé.txt", f.filename)
        assert_equal("text/plain", f.content_type)
        assert_equal("cöntent".encode("utf-8"), f.read())

    @test
    def multipart_put_request_with_file(self) -> None:
        self.env["REQUEST_METHOD"] = "PUT"
        self.setup_multipart_file_request(
            "my-file", "my-file.txt", "content", "text/plain"
        )
        args = parse_args(
            self.env, [("my-file", "file", Multiplicity.REQUIRED)]
        )
        assert_in("my-file", args)
        f = args["my-file"]
        assert_equal("my-file.txt", f.filename)
        assert_equal("text/plain", f.content_type)
        assert_equal(b"content", f.read())

    @test
    def read_file_as_value(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.setup_multipart_file_request(
            "foo", "my-file.txt", "123", "text/plain"
        )
        args = parse_args(self.env, [("foo", int, Multiplicity.REQUIRED)])
        assert_equal(123, args["foo"])

    @test
    def read_value_as_file(self) -> None:
        self.env["REQUEST_METHOD"] = "GET"
        self.add_path_argument("foo", "bar")
        args = parse_args(self.env, [("foo", "file", Multiplicity.REQUIRED)])
        assert_in("foo", args)
        f = args["foo"]
        assert_equal("", f.filename)
        assert_equal("application/octet-stream", f.content_type)
        assert_equal(b"bar", f.read())

    @test
    def read_value_as_file_with_umlauts(self) -> None:
        self.env["REQUEST_METHOD"] = "GET"
        self.add_path_argument("foo", "bär")
        args = parse_args(self.env, [("foo", "file", Multiplicity.REQUIRED)])
        assert_in("foo", args)
        f = args["foo"]
        assert_equal("", f.filename)
        assert_equal("application/octet-stream", f.content_type)
        assert_equal("bär".encode("utf-8"), f.read())

    @test
    def post_wrong_content_type__optional_args(self) -> None:
        """This exposes a bug in Python's cgi module that will raise a
        TypeError when no request string was provided. See
        <https://bugs.python.org/issue32029>.
        """
        self.env["REQUEST_METHOD"] = "POST"
        self.env["CONTENT_TYPE"] = "application/octet-stream"
        self.env["CONTENT_LENGTH"] = "2"
        self.env["wsgi.input"] = BytesIO(b"AB")
        args = parse_args(self.env, [("foo", str, Multiplicity.OPTIONAL)])
        assert_equal({}, args)

    @test
    def post_wrong_content_type__required_args(self) -> None:
        self.env["REQUEST_METHOD"] = "POST"
        self.env["CONTENT_TYPE"] = "application/octet-stream"
        self.env["CONTENT_LENGTH"] = "2"
        self.env["wsgi.input"] = BytesIO(b"AB")
        with assert_raises(BadRequest):
            parse_args(self.env, [("foo", str, Multiplicity.REQUIRED)])

    @test
    def patch_wrong_content_type__optional_args(self) -> None:
        self.env["REQUEST_METHOD"] = "PATCH"
        self.env["CONTENT_TYPE"] = "application/octet-stream"
        self.env["CONTENT_LENGTH"] = "2"
        self.env["wsgi.input"] = BytesIO(b"AB")
        args = parse_args(self.env, [("foo", str, Multiplicity.OPTIONAL)])
        assert_equal({}, args)

    @test
    def patch_wrong_content_type__required_args(self) -> None:
        self.env["REQUEST_METHOD"] = "PATCH"
        self.env["CONTENT_TYPE"] = "application/octet-stream"
        self.env["CONTENT_LENGTH"] = "2"
        self.env["wsgi.input"] = BytesIO(b"AB")
        with assert_raises(BadRequest):
            parse_args(self.env, [("foo", str, Multiplicity.REQUIRED)])

    @test
    def no_exhaustive_check(self) -> None:
        self.add_path_argument("foo", "v")
        self.add_path_argument("unknown", "v")
        with assert_succeeds(ArgumentsError):
            parse_args(self.env, [("foo", str, Multiplicity.OPTIONAL)])

    @test
    def exhaustive_check__succeeds(self) -> None:
        self.add_path_argument("foo", "v")
        with assert_succeeds(ArgumentsError):
            parse_args(
                self.env,
                [("foo", str, Multiplicity.OPTIONAL)],
                exhaustive=True,
            )

    @test
    def exhaustive_check__fails(self) -> None:
        self.add_path_argument("foo", "v")
        self.add_path_argument("unknown", "v")
        try:
            parse_args(
                self.env,
                [("foo", str, Multiplicity.OPTIONAL)],
                exhaustive=True,
            )
        except ArgumentsError as exc:
            assert_equal({"unknown": "unknown argument"}, exc.arguments)
        else:
            fail("ArgumentsError not raised")

    @test
    def unsupported_method(self) -> None:
        self.env["REQUEST_METHOD"] = "UNKNOWN"
        with assert_raises(ValueError):
            parse_args(self.env, [])


class ArgumentParserTest(TestCase):
    @test
    def parse_args__post_twice(self) -> None:
        environ = {
            "wsgi.input": BytesIO(b"foo=bar&abc=def"),
            "REQUEST_METHOD": "POST",
            "CONTENT_LENGTH": "15",
            "CONTENT_TYPE": "application/x-www-form-urlencoded",
        }
        parser = ArgumentParser(environ)
        args1 = parser.parse_args([("foo", str, Multiplicity.REQUIRED)])
        assert_equal({"foo": "bar"}, args1)
        args2 = parser.parse_args(
            [
                ("foo", str, Multiplicity.REQUIRED),
                ("abc", str, Multiplicity.REQUIRED),
            ]
        )
        assert_equal({"foo": "bar", "abc": "def"}, args2)

    @test
    def exhaustive_with_previous_calls(self) -> None:
        environ = {
            "wsgi.input": BytesIO(b"foo=bar&abc=def"),
            "REQUEST_METHOD": "POST",
            "CONTENT_LENGTH": "15",
            "CONTENT_TYPE": "application/x-www-form-urlencoded",
        }
        parser = ArgumentParser(environ)
        parser.parse_args(
            [("foo", str, Multiplicity.REQUIRED)], exhaustive=False
        )
        with assert_succeeds(ArgumentsError):
            parser.parse_args(
                [("abc", str, Multiplicity.REQUIRED)], exhaustive=True
            )



from __future__ import annotations

from collections.abc import Iterable
from http import HTTPStatus
from io import BytesIO

import pytest
from werkzeug.exceptions import UnsupportedMediaType
from werkzeug.wrappers import Request

from rouver.args import Multiplicity
from rouver.handler import RouteHandlerBase
from rouver_test.testutil import StubStartResponse, default_environment


class StubHandler(RouteHandlerBase):
    response: Iterable[bytes] = []

    def prepare_response(self) -> Iterable[bytes]:
        self.respond()
        return self.response


class TestRouteHandlerBase:
    @pytest.fixture(autouse=True)
    def setup_environment(self) -> None:
        self.environ = default_environment()
        self.start_response = StubStartResponse()

    def call_handler(self, handler_class: type[RouteHandlerBase]) -> bytes:
        handler = handler_class(self.environ, self.start_response)
        response = b"".join(handler)
        handler.close()
        return response

    def test_attributes(self) -> None:
        handler = StubHandler(self.environ, self.start_response)
        assert handler.request.environ is self.environ
        assert isinstance(handler.request, Request)
        assert handler.start_response is self.start_response

    def test_path_args__from_environment(self) -> None:
        self.environ["rouver.path_args"] = ["foo"]
        handler = StubHandler(self.environ, self.start_response)
        assert handler.path_args == ["foo"]

    def test_path_args__default(self) -> None:
        handler = StubHandler(self.environ, self.start_response)
        assert handler.path_args == []

    def test_path_args__not_a_list(self) -> None:
        self.environ["rouver.path_args"] = "not-a-list"
        handler = StubHandler(self.environ, self.start_response)
        assert handler.path_args == []

    def test_wildcard_path__from_environment(self) -> None:
        self.environ["rouver.wildcard_path"] = "/foo/bar"
        handler = StubHandler(self.environ, self.start_response)
        assert handler.wildcard_path == "/foo/bar"

    def test_wildcard_path__decode(self) -> None:
        self.environ["rouver.wildcard_path"] = "/foo%2Fb%C3%A4r"
        handler = StubHandler(self.environ, self.start_response)
        assert handler.wildcard_path == "/foo/bär"

    def test_wildcard_path__decode_errors(self) -> None:
        self.environ["rouver.wildcard_path"] = "/foo%2Fb%C3r"
        handler = StubHandler(self.environ, self.start_response)
        assert handler.wildcard_path == "/foo/b�r"

    def test_wildcard_path__default(self) -> None:
        handler = StubHandler(self.environ, self.start_response)
        assert handler.wildcard_path == ""

    def test_wildcard_path__not_a_string(self) -> None:
        self.environ["rouver.wildcard_path"] = b"not-a-str"
        handler = StubHandler(self.environ, self.start_response)
        assert handler.wildcard_path == ""

    def test_parse_args__post_twice(self) -> None:
        self.environ["wsgi.input"] = BytesIO(b"foo=bar&abc=def")
        self.environ["REQUEST_METHOD"] = "POST"
        self.environ["CONTENT_LENGTH"] = "15"
        self.environ["CONTENT_TYPE"] = "application/x-www-form-urlencoded"
        handler = StubHandler(self.environ, self.start_response)
        args1 = handler.parse_args([("foo", str, Multiplicity.REQUIRED)])
        assert args1 == {"foo": "bar"}
        args2 = handler.parse_args(
            [
                ("foo", str, Multiplicity.REQUIRED),
                ("abc", str, Multiplicity.REQUIRED),
            ]
        )
        assert args2 == {"foo": "bar", "abc": "def"}

    def test_parse_args__works_in_response_handler(self) -> None:
        class MyHandler(RouteHandlerBase):
            def prepare_response(self) -> Iterable[bytes]:
                self.parse_args([])
                return self.respond()

        self.environ["wsgi.input"] = BytesIO(b"foo=bar&abc=def")
        self.environ["REQUEST_METHOD"] = "POST"
        self.environ["CONTENT_LENGTH"] = "15"
        self.environ["CONTENT_TYPE"] = "application/x-www-form-urlencoded"
        handler = MyHandler(self.environ, self.start_response)
        iter(handler)

    def test_parse_json_request__default_encoding(self) -> None:
        self.environ["wsgi.input"] = BytesIO(b'{ "f\xc3\xb6o": 42 }')
        self.environ["CONTENT_LENGTH"] = "14"
        self.environ["CONTENT_TYPE"] = "application/json"
        handler = StubHandler(self.environ, self.start_response)
        j = handler.parse_json_request()
        assert j == {"föo": 42}

    def test_parse_json_request__explicit_encoding(self) -> None:
        self.environ["wsgi.input"] = BytesIO(b'{ "f\xf6o": 42 }')
        self.environ["CONTENT_LENGTH"] = "13"
        self.environ["CONTENT_TYPE"] = "application/json; charset=iso-8859-1"
        handler = StubHandler(self.environ, self.start_response)
        j = handler.parse_json_request()
        assert j == {"föo": 42}

    def test_parse_json_request__unknown_encoding(self) -> None:
        self.environ["wsgi.input"] = BytesIO(b"{}")
        self.environ["CONTENT_LENGTH"] = "2"
        self.environ["CONTENT_TYPE"] = "application/json; charset=unknown"
        handler = StubHandler(self.environ, self.start_response)
        with pytest.raises(UnsupportedMediaType):
            handler.parse_json_request()

    def test_parse_json_request__no_content_type(self) -> None:
        self.environ["wsgi.input"] = BytesIO(b"{}")
        self.environ["CONTENT_LENGTH"] = "2"
        handler = StubHandler(self.environ, self.start_response)
        with pytest.raises(UnsupportedMediaType):
            handler.parse_json_request()

    def test_parse_json_request__wrong_content_type(self) -> None:
        self.environ["wsgi.input"] = BytesIO(b"{}")
        self.environ["CONTENT_LENGTH"] = "2"
        self.environ["CONTENT_TYPE"] = "application/octet-stream"
        handler = StubHandler(self.environ, self.start_response)
        with pytest.raises(UnsupportedMediaType):
            handler.parse_json_request()

    def test_parse_json_request__invalid_data(self) -> None:
        self.environ["wsgi.input"] = BytesIO(b"INVALID")
        self.environ["CONTENT_LENGTH"] = "7"
        self.environ["CONTENT_TYPE"] = "application/json"
        handler = StubHandler(self.environ, self.start_response)
        with pytest.raises(UnsupportedMediaType):
            handler.parse_json_request()

    def test_respond(self) -> None:
        StubHandler.response = [b"foo", b"bar"]
        response = self.call_handler(StubHandler)
        self.start_response.assert_status(HTTPStatus.OK)
        assert response == b"foobar"

    def test_closeable_response(self) -> None:
        StubHandler.response = BytesIO(b"foo")
        response = self.call_handler(StubHandler)
        self.start_response.assert_status(HTTPStatus.OK)
        assert response == b"foo"
        assert StubHandler.response.closed



from __future__ import annotations

import re
from collections.abc import Iterable
from http import HTTPStatus
from io import BytesIO, StringIO
from typing import Any

from asserts import assert_equal, assert_false, assert_regex, assert_true

from rouver.types import Header, StartResponseReturnType, WSGIEnvironment

_status_re = re.compile(r"^(\d\d\d) (.*)$")


class StubStartResponse:
    def __init__(self) -> None:
        self.was_called = False
        self.status = ""
        self.headers: list[Header] = []

    def __call__(
        self, status: str, headers: Iterable[Header], exc_info: Any = None
    ) -> StartResponseReturnType:
        assert_false(self.was_called, "start_response() called twice")
        assert_regex(status, _status_re)
        self.was_called = True
        self.status = status
        self.headers = list(headers)
        return lambda s: None

    @property
    def status_code(self) -> int:
        self.assert_was_called()
        assert len(self.status) >= 3
        return int(self.status[:3])

    def assert_was_called(self) -> None:
        assert_true(self.was_called, "start_response() was not called")

    def assert_status(self, status: HTTPStatus) -> None:
        assert_equal(status.value, self.status_code)

    def assert_header_missing(self, name: str) -> None:
        value = self._find_header(name)
        if value is not None:
            raise AssertionError("header {} unexpectedly found".format(name))

    def assert_header_equals(self, name: str, value: str) -> None:
        header_value = self._find_header(name)
        if header_value is None:
            raise AssertionError("missing header '{}'".format(name))
        assert_equal(
            value,
            header_value,
            "'{}': '{}' != '{}".format(name, value, header_value),
        )

    def _find_header(self, name: str) -> str | None:
        self.assert_was_called()
        found = None
        for header_name, header_value in self.headers:
            if header_name.lower() == name.lower():
                if not isinstance(header_value, str):
                    raise AssertionError("invalue header value")
                if found is not None:
                    raise AssertionError(
                        "duplicate header '{}'".format(header_name)
                    )
                found = header_value
        return found


def default_environment() -> WSGIEnvironment:
    return {
        "REQUEST_METHOD": "GET",
        "SERVER_NAME": "www.example.com",
        "SERVER_PORT": "80",
        "SERVER_PROTOCOL": "HTTP/1.1",
        "wsgi.version": (1, 0),
        "wsgi.url_scheme": "http",
        "wsgi.input": BytesIO(b""),
        "wsgi.errors": StringIO(),
        "wsgi.multithread": False,
        "wsgi.multiprocess": False,
        "wsgi.run_once": False,
    }
